;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; classes.lisp -- persistent objects via metaobjects
;;; 
;;; Initial version 8/26/2004 by Andrew Blumberg
;;; <ablumberg@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;


(in-package "ELEPHANT")

(defmethod initialize-instance :before  ((instance persistent)
					 &rest initargs
					 &key from-oid)
  "Sets the OID."
  (declare (ignore initargs))
  (if (not from-oid)
      (setf (oid instance) (next-oid *store-controller*))
      (setf (oid instance) from-oid))
  (cache-instance *store-controller* instance))

(defclass persistent-object (persistent)
  ((%persistent-slots :transient t))
  (:documentation 
"Superclass of all user-defined persistent classes.  This is
automatically inherited if you use the persistent-metaclass
metaclass.")
  (:metaclass persistent-metaclass))

(defmethod shared-initialize :around ((class persistent-metaclass) slot-names &rest args &key direct-superclasses)
  "Ensures we inherit from persistent-object."
  (let* ((persistent-metaclass (find-class 'persistent-metaclass))
	 (persistent-object (find-class 'persistent-object))
	 (not-already-persistent (loop for superclass in direct-superclasses
				       never (eq (class-of superclass) persistent-metaclass))))
    (if (and (not (eq class persistent-object)) not-already-persistent)
	(apply #'call-next-method class slot-names
	       :direct-superclasses (cons persistent-object
					  direct-superclasses) args)
	(call-next-method))))

(defmethod shared-initialize :around ((instance persistent-object) slot-names &rest initargs &key &allow-other-keys)
  "Initializes the persistent slots via initargs or forms.
This seems to be necessary because it is typical for
implementations to optimize setting the slots via initforms
and initargs in such a way that slot-value-using-class et al
aren't used.  Calls the next method for the transient slots."
  (let* ((class (class-of instance))
	 (persistent-slot-names (persistent-slot-names class)))
    (flet ((persistent-slot-p (item) 
	     (member item persistent-slot-names :test #'eq)))
      (let ((transient-slot-inits 
	     (if (eq slot-names t) ; t means all slots
		 (transient-slot-names class)
		 (remove-if #'persistent-slot-p slot-names)))
	    (persistent-slot-inits
	     (if (eq slot-names t) persistent-slot-names
		 (remove-if-not #'persistent-slot-p slot-names))))
	;; initialize the persistent slots
	(flet ((initialize-from-initarg (slot-def)
		 (loop for initarg in initargs
		       with slot-initargs = (slot-definition-initargs slot-def)
		       when (member initarg slot-initargs :test #'eq)
		       do 
		       (setf (slot-value-using-class class instance slot-def) 
			     (getf initargs initarg))
		       (return t))))
	  (loop for slot-def in (class-slots class)
		unless (initialize-from-initarg slot-def)
		when (member (slot-definition-name slot-def) persistent-slot-inits :test #'eq)
		unless (slot-boundp-using-class class instance slot-def)
		do
		(let ((initfun (slot-definition-initfunction slot-def)))
		  (when initfun
		    (setf (slot-value-using-class class instance slot-def)
			  (funcall initfun))))))
	;; let the implementation initialize the transient slots
	(apply #'call-next-method instance transient-slot-inits initargs)))))

(defmethod update-instance-for-different-class :around ((previous persistent) (current persistent) &rest initargs &key)
  "Need to also update the persistent-slots, which have
:class allocation."
  (let ((new-persistent-slots 
	 (loop for slotd in (class-slots (class-of current))
	       for slot-name = (slot-definition-name slotd)
	       with old-slot-names = (mapcar #'slot-definition-name
					     (class-slots (class-of previous)))
	       when (and (not (member slot-name old-slot-names :test #'eq))
			 (persistent-p slotd))
	       collect slot-name)))
    (apply #'shared-initialize current new-persistent-slots initargs)
    (call-next-method)))

(defmethod slot-value-using-class :around (class (instance persistent-object) (slot-def persistent-slot-definition))
  "Get the slot value from the database."
  (declare (optimize (speed 3))
	   (ignore class))
  (let ((name (slot-definition-name slot-def)))
    (persistent-slot-reader instance name)))

(defmethod (setf slot-value-using-class) :around (new-value class (instance persistent-object) (slot-def persistent-slot-definition))
  "Set the slot value in the database."
  (declare (optimize (speed 3))
	   (ignore class))
  (let ((name (slot-definition-name slot-def)))
    (persistent-slot-writer new-value instance name)))

(defmethod slot-boundp-using-class :around (class (instance persistent-object) (slot-def persistent-slot-definition))
  "Checks if the slot exists in the database."
  (declare (optimize (speed 3))
	   (ignore class))
  (let ((name (slot-definition-name slot-def)))
    (persistent-slot-boundp instance name)))

(defmethod slot-makunbound-using-class :around (class (instance persistent-object) (slot-def persistent-slot-definition))
  "Deletes the slot from the database."
  (declare (optimize (speed 3))
	   (ignore class))
  (with-buffer-streams (key-buf)
    (buffer-write-int (oid instance) key-buf)
    (serialize (slot-definition-name slot-def) key-buf)
    (db-delete-buffered
     (controller-db *store-controller*) key-buf
     :transaction *current-transaction*
     :auto-commit *auto-commit*))
  instance)
  