;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; controller.lisp -- Lisp interface to a Berkeley DB store
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;

(in-package "ELEPHANT")

(defclass store-controller ()  
  ((path :type (or pathname string)
	 :accessor controller-path
	 :initarg :path)
   (environment :type (or null pointer-void) 
		:accessor controller-environment)
   (db :type (or null pointer-void) :accessor controller-db)
   (btrees :type (or null pointer-void) :accessor controller-btrees)
   (indices :type (or null pointer-void) :accessor controller-indices)
   (indices-assoc :type (or null pointer-void) 
		  :accessor controller-indices-assoc)
   (root :reader controller-root)
   (instance-cache :accessor instance-cache
		   :initform (make-cache-table :test 'eql)))
  (:documentation "Class of objects responsible for the
book-keeping of holding DB handles, the cache, table
creation, counters, locks, the root (for garbage collection,)
et cetera."))

(defgeneric open-controller (sc &key recover recover-fatal thread)
  (:documentation 
   "Opens the underlying environment and all the necessary
database tables."))

(defgeneric close-controller (sc)
  (:documentation 
   "Close the db handles and environment.  Tries to wipe out
references to the db handles."))

(defun add-to-root (key value &key (store-controller *store-controller*))
  "Add an arbitrary persistent thing to the root, so you can
retrieve it in a later session.  N.B. this means it (and
everything it points to) won't get gc'd."
  (declare (type store-controller store-controller))
  (setf (get-value key (controller-root store-controller)) value))

(defun get-from-root (key &key (store-controller *store-controller*))
  "Get a something from the root."
  (declare (type store-controller store-controller))
  (get-value key (controller-root store-controller)))

(defun remove-from-root (key &key (store-controller *store-controller*))
  "Remove something from the root."
  (declare (type store-controller store-controller))
  (remove-kv key (controller-root store-controller)))

(defun cache-instance (sc obj)
  "Cache a persistent object with the controller."
  (declare (type store-controller sc))
  (setf (get-cache (oid obj) (instance-cache sc)) obj))

(defun get-cached-instance (sc oid class-name)
  "Get a cached instance, or instantiate!"
  (declare (type store-controller sc)
	   (type fixnum oid))
  (let ((obj (get-cache oid (instance-cache sc))))
    (if obj obj
	;; Should get cached since make-instance calls cache-instance
	(make-instance class-name :from-oid oid))))

;; OID stuff
;; This stuff is all a hack until sequences appear in Sleepycat 4.3
(defvar %oid-entry (uffi:allocate-foreign-object :char 12))
(defvar %oid-lock (uffi:allocate-foreign-object :char 16))

(eval-when (:load-toplevel)
  (loop for c across "%ELEPHANTOID"
	for i from 0 to 11
	do (setf (uffi:deref-array %oid-entry '(:array :char) i)
		 (char-code c)))
  (loop for c across "%ELEPHANTOIDLOCK"
	for i from 0 to 15
	do (setf (uffi:deref-array %oid-lock '(:array :char) i) 
		 (char-code c)))
  )

(defvar %oid-entry-length 12)
(defvar %oid-lock-length 16)

(defun next-oid (sc)
  "Get the next OID."
  (declare (type store-controller sc))
  (sleepycat::next-counter (controller-environment sc)
			   (controller-db sc)
			   *current-transaction*
			   %oid-entry %oid-entry-length
			   %oid-lock %oid-lock-length))

;; Open/close     
(defmethod open-controller ((sc store-controller) &key (recover nil)
			    (recover-fatal nil) (thread t))
  (let ((env (db-env-create)))
    ;; thread stuff?
    (setf (controller-environment sc) env)
    (db-env-open env (controller-path sc) :create t :init-txn t :init-lock t 
		 :init-mpool t :init-log t :thread thread
		 :recover recover :recover-fatal recover-fatal)
    (let ((db (db-create env))
	  (btrees (db-create env))
	  (indices (db-create env))
	  (indices-assoc (db-create env)))
      (setf (controller-db sc) db)
      (db-open db :file "%ELEPHANT" :database "%ELEPHANTDB" 
	       :auto-commit t :type DB-BTREE :create t :thread thread)

      (setf (controller-btrees sc) btrees)
      (sleepycat::db-set-lisp-compare btrees)
      (db-open btrees :file "%ELEPHANT" :database "%ELEPHANTBTREES" 
	       :auto-commit t :type DB-BTREE :create t :thread thread)

      (setf (controller-indices sc) indices)
      (sleepycat::db-set-lisp-compare indices)
      (sleepycat::db-set-lisp-dup-compare indices)
      (db-set-flags indices :dup-sort t)
      (db-open indices :file "%ELEPHANT" :database "%ELEPHANTINDICES" 
	       :auto-commit t :type DB-BTREE :create t :thread thread)

      (setf (controller-indices-assoc sc) indices-assoc)
      (sleepycat::db-set-lisp-compare indices-assoc)
      (sleepycat::db-set-lisp-dup-compare indices-assoc)
      (db-set-flags indices-assoc :dup-sort t)
      (db-open indices-assoc :file "%ELEPHANT" :database "%ELEPHANTINDICES" 
	       :auto-commit t :type DB-UNKNOWN :thread thread :rdonly t)
      (sleepycat::db-fake-associate btrees indices-assoc :auto-commit t)

      (let ((root (make-instance 'btree :from-oid -1)))
	(setf (slot-value sc 'root) root)
	(with-transaction ()
	  (with-buffer-streams (key-buf value-buf)
	    (let ((key-b (buffer-stream-buffer key-buf)))
	      (setf (buffer-stream-buffer key-buf) %oid-entry)
	      (setf (sleepycat::buffer-stream-size key-buf) %oid-entry-length)
	      (unless (db-get-key-buffered db key-buf value-buf)
		(reset-buffer-stream value-buf)
		(buffer-write-int 0 value-buf)
		(db-put-buffered db key-buf value-buf))
	      (setf (buffer-stream-buffer key-buf) key-b))))
	sc))))

(defmethod close-controller ((sc store-controller))
  (when (slot-value sc 'root)
    ;; no root
    (setf (slot-value sc 'root) nil)
    ;; clean instance cache
    (setf (instance-cache sc) (make-cache-table :test 'eql))
    ;; close handles / environment
    (db-close (controller-indices-assoc sc))
    (setf (controller-indices-assoc sc) nil)
    (db-close (controller-indices sc))
    (setf (controller-indices sc) nil)
    (db-close (controller-btrees sc))
    (setf (controller-btrees sc) nil)
    (db-close (controller-db sc))
    (setf (controller-db sc) nil)
    (db-env-close (controller-environment sc))
    (setf (controller-environment sc) nil)
    nil))

(defmacro with-open-controller ((&optional (sc '*store-controller*))
				&body body)
  "Executes body with the specified controller open, closing
the controller unconditionally on exit."
  `(unwind-protect
       (progn
	 (let (*store-controller* (open-controller ,sc))
	   (declare (special *store-controller*))
	   ,@body))
     (close-controller ,sc)))

(defun open-store (path  &key (recover nil)
		   (recover-fatal nil) (thread t))
  "Conveniently open a store controller."
  (setq *store-controller* (make-instance 'store-controller :path path))
  (open-controller *store-controller* :recover recover 
		   :recover-fatal recover-fatal :thread thread))

(defun close-store ()
  "Conveniently close the store controller."
  (close-controller *store-controller*))

(defmacro with-open-store ((path) &body body)
  "Executes the body with an open controller,
unconditionally closing the controller on exit."
  `(let ((*store-controller* (make-instance 'store-controller :path ,path)))
    (declare (special *store-controller*))
    (open-controller *store-controller*)
    (unwind-protect
	 (progn ,@body)
      (close-controller *store-controller*))))

(defun start-transaction (&key (parent *current-transaction*))
  "Start a transaction.  May be nested but not interleaved."
  (vector-push-extend *current-transaction* *transaction-stack*)
  (setq *current-transaction* 
	(db-transaction-begin (controller-environment *store-controller*)
			      :parent parent)))

(defun commit-transaction ()
  "Commit the current transaction."
  (db-transaction-commit)
  (setq *current-transaction* (vector-pop *transaction-stack*)))

(defun abort-transaction ()
  "Abort the current transaction."
  (db-transaction-abort)
  (setq *current-transaction* (vector-pop *transaction-stack*)))
